/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>
#include <poll.h>
#include <inttypes.h>
#include <libgen.h>
#include <libwrtd.h>

static void help()
{
	fprintf(stderr, "wrtd-logging -D 0x<hex-number>\n");
	fprintf(stderr, "It shows logging information coming from Real-Time applications\n");
	fprintf(stderr, "-D device id\n");
	fprintf(stderr, "-n number of messages to read (0 means infinite)\n");
	fprintf(stderr, "-s show current logging level for all channels\n");
	exit(1);
}


static void print_logging (struct wrtd_node *wrtd, int n_read)
{
	struct wrtd_log_entry log;
	int i, count, valid, cpu;
	uint64_t ps;

	valid = wrtd_in_is_valid(wrtd);
	if (!valid) {
		fprintf(stderr,
			"Cannot run logging, invalid input firmware: %s\n",
			wrtd_strerror(errno));
		return;
	}
	valid = wrtd_out_is_valid(wrtd);
	if (!valid) {
		fprintf(stderr,
			"Cannot run logging, invalid output firmware: %s\n",
			wrtd_strerror(errno));
		return;
	}


	for(i = 0; i < n_read || n_read == 0; i++) {
		count = wrtd_log_read(wrtd, &log, 1, -1);
		if (count <= 0) {
			fprintf(stderr, "Cannot read message: %s\n",
				wrtd_strerror(errno));
			break;
		}
		cpu = (log.type & WRTD_LOG_CPU_MASK) >> WRTD_LOG_CPU_SHIFT;
		fprintf(stdout, "Device      %s\n",
			cpu == WRTD_CPU_TDC ? "tdc" : "fd");
		fprintf(stdout, "Event Type  %s\n", wrtd_strlogging(log.type));
		if (log.type == WRTD_LOG_PROMISC || log.channel < 0)
			fprintf(stdout, "Channel     --\n");
		else
			fprintf(stdout, "Channel     %d\n", log.channel);
		fprintf(stdout, "Miss reason %x\n", log.miss_reason);
		fprintf(stdout, "Seq         %d\n", log.seq);
		if (log.type == WRTD_LOG_RAW)
			fprintf(stdout, "Identifier  ----:----:----\n");
		else
			fprintf(stdout, "Identifier  %04x:%04x:%04x\n",
				log.id.system, log.id.source_port, log.id.trigger);

		ps = (uint64_t) log.ts.ticks * 8000 + (uint64_t)log.ts.frac * 8000ULL / 4096ULL;

		fprintf(stdout,
			"Timestamp   %"PRIu64"s  %"PRIu32"tick %"PRIu32"frac "
			"(%03d,%03d,%03dns + %03dps)\n",
			log.ts.seconds, log.ts.ticks, log.ts.frac,
			(int)(ps / (1000 * 1000 * 1000ULL)),
			(int)(ps / (1000 * 1000ULL)) % 1000,
			(int)(ps / 1000ULL) % 1000,
			(int)(ps % 1000ULL));
		fprintf(stdout, "----\n");
                fflush(stdout);

		i++;
	}
}

static void show_logging_level(struct wrtd_node *dev, int cpu)
{
	uint32_t log_level;
	char log_level_str[128];
	int i, err, max = cpu == WRTD_CPU_FD ? FD_NUM_CHANNELS : TDC_NUM_CHANNELS;

	fprintf(stdout, "%s log levels\n",
		cpu == WRTD_CPU_FD ? "Output (FD)" : "Input (TDC)");
	for (i = 0; i < max; i++) {
		if (cpu == WRTD_CPU_FD)
			err = wrtd_out_log_level_get(dev, i, &log_level);
		else
			err = wrtd_in_log_level_get(dev, i, &log_level);
		if (err) {
			fprintf(stdout, "\tchannel %d: --- ERROR ---\n", i);
		} else {
			wrtd_strlogging_full(log_level_str, log_level);
			fprintf(stdout, "\tchannel %d: %s\n", i, log_level_str);
		}
	}
}


#define N_LOG 2
int main(int argc, char *argv[])
{
	struct wrtd_node *wrtd;
	int n_read = 0;
	char *endptr;
	int err, show_log = 0;
	uint32_t dev_id = 0;
	char c;

	while ((c = getopt (argc, argv, "hD:n:s")) != -1) {
		switch (c) {
		default:
			help();
			break;
		case 'D':
			dev_id = strtoul(optarg, &endptr, 0);
			if (*endptr != 0) {
				fprintf(stderr, "bad value for -D");
				exit (1);
			}
			break;
		case 'n':
			n_read = strtoul(optarg, &endptr, 0);
			if (*endptr != 0) {
				fprintf(stderr, "bad value for -n");
				exit (1);
			}
			break;
		case 's':
			show_log = 1;
			break;
		}
	}

	if (dev_id == 0)
		help();

	atexit(wrtd_exit);
	err = wrtd_init();
	if (err) {
		fprintf(stderr, "Cannot init White Rabbit Node Core lib: %s\n",
			trtl_strerror(errno));
		exit(1);
	}

	wrtd = wrtd_open(dev_id);
	if (!wrtd) {
		fprintf(stderr, "Cannot open WRNC: %s\n", wrtd_strerror(errno));
		exit(1);
	}

	if (show_log) {
		show_logging_level(wrtd, WRTD_CPU_TDC);
		show_logging_level(wrtd, WRTD_CPU_FD);
		exit(0);
	}

	print_logging (wrtd, n_read);

	wrtd_close(wrtd);
	exit(0);
}
